from test_junkie.constants import TestCategory


class ReportTemplate:

    LABEL_COLOR = "#aebac4"

    @staticmethod
    def get_body_template():
        return """
                <!DOCTYPE html>
                <html lang="en">
                    <head>
                        <meta charset="UTF-8">
                        <title>Test Junkie HTML Report</title>
                        <script src="https://code.jquery.com/jquery-3.3.1.min.js"
                                integrity="sha256-FgpCb/KJQlLNfOu91ta32o/NMZxltwRo8QtmkMRdAu8="
                                crossorigin="anonymous"></script>
                        <script src="https://cdnjs.cloudflare.com/ajax/libs/Chart.js/2.7.3/Chart.bundle.min.js"></script>
                        <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/materialize/0.100.2/css/materialize.min.css">
                        <script src="https://cdnjs.cloudflare.com/ajax/libs/materialize/0.100.2/js/materialize.min.js"></script>
                        <link href="https://fonts.googleapis.com/icon?family=Material+Icons" rel="stylesheet">
                        <link href="https://cdn.datatables.net/1.10.19/css/jquery.dataTables.min.css" rel="stylesheet">
                        <script src="https://www.amcharts.com/lib/4/core.js"></script>
                        <script src="https://www.amcharts.com/lib/4/charts.js"></script>
                        <script src="https://www.amcharts.com/lib/4/themes/dark.js"></script>
                        <script src="https://www.amcharts.com/lib/4/themes/animated.js"></script>
                        <script src="https://cdn.datatables.net/1.10.19/js/jquery.dataTables.min.js"></script>
                    </head>
                    <body style="background-color: #3d4f58;">
                        <style>
                            .absolute_cards{{
                                min-height: 170px;
                                overflow: hidden;
                            }}

                            .absolute_value{{
                                font-size: 27px;
                            }}

                            .card:hover {{
                                background: #193746 !important;
                                transition: all .3s ease-in;
                            }}
                            
                            body {{
                                font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Helvetica, Arial, sans-serif, "Apple Color Emoji", "Segoe UI Emoji", "Segoe UI Symbol";
                            }}
                            
                            .success {{
                                background: #12d479;
                            }}
                            .fail {{
                                background: #fcd75f;
                            }}
                            .error {{
                                background: #ff7651;
                            }}
                            .ignore {{
                                background: #cce4eb;
                            }}
                            .skip {{
                                background: #34bff5;
                            }}
                            .cancel {{
                                background: #f19def;
                            }}
                            .info-badge {{
                                background: #aac7d2;
                            }}
                            
                            .parent-badge{{
                                margin-top: 8px;
                            }}
                            
                            .badge {{
                                border-radius: 3px;
                                margin-right: 7px;
                                color: #333 !important;
                            }}
                            
                            .traceback {{
                                background: #23323a;
                                border-radius: 3px;
                                padding: 15px;
                                color: #fd5858;
                            }}
                            
                            .collapsible-header {{
                                background: #37474f !important;
                                border-color: #193746;
                            }}
                            
                            .collapsible.expandable {{
                                border: 1px #37474f;
                            }}
                            
                            .collapsible-header:hover {{
                                background: #193746 !important;
                                transition: all .3s ease-in;
                                transition: all .3s ease-out;
                            }}
                            
                            .collapsible-body {{
                                border-color: transparent;
                            }}
                            
                            .attempt-header {{
                                padding: 5px;
                                padding-bottom: 0px;
                                border: 0px;
                            }}
                            
                            .parameter {{
                              white-space: nowrap;
                              overflow: hidden;
                              text-overflow: ellipsis;
                              font-size: 14px;
                            }}
                        </style>
                        {body}
                        <script>var database_lol = {database_lol}</script>
                    </body>
                </html> 
               """

    @staticmethod
    def get_donation_options():

        return """
                <div class="fixed-action-btn horizontal">
                    <a class="btn-floating btn-large waves-effect waves-light orange darken-4 modal-trigger" href="#developerModal"><i class="material-icons">favorite</i></a>
                </div>
                <script>
                    $(document).ready(function(){
                        $('#developerModal').modal();
                    });
                </script>
                <div id="developerModal" class="modal">
                    <div class="modal-content" style="color: black;">
                      <h4>Greetings!</h4>
                        <p>Some features are not yet finished. 
                           I'm <a href="https://www.linkedin.com/in/arturspirin/" target="_blank">Artur</a>
                           by the way, the developer of
                           <a href="https://github.com/ArturSpirin/test_junkie" target="_blank">Test Junkie</a></p>
                        <p>It took a lot of effort, all my free time outside my day job, and many sleepless
                            nights to get this project to this point, that includes styling and doing data
                            processing in order to render all those charts and tables. I have been motivated by
                            the idea that it may become useful for testers, QA managers, QA team leads, and
                            even project managers and developers. Thus I open sourced this project initially
                            knowing that I wont make much (if any at all) money by doing it.</p>
                        <p>Ask yourself if all the effort I put in and the benefits Test Junkie brings to you
                            and/or your company (if any) are worth a donation. This donation will keep me
                            motivated and directly support this project.</p>
                        <p>Ways you can donate:</p>
                        <ol>
                            <li>Share Test Junkie on social media, with your friend and/or colleagues.
                                Tell them how it helped you in your project.</li>
                            <li>Write me a message on <a href="https://www.linkedin.com/in/arturspirin/"
                                                         target="_blank">Linked In</a> and tell me about your
                                experience with Test Junkie.</li>
                            <li>If you can afford, you can donate through
                                <a href="https://www.paypal.com/cgi-bin/webscr?cmd=_donations&business=FJPYWX5B776YS
                                &currency_code=USD&source=url" target="_blank">PayPal</a>.
                                No amount is small amount, its the thought that counts!</li>
                            <li>You can also back me on
                                <a href="https://www.patreon.com/join/arturspirin/" target="_blank">Patreon</a>
                                on behalf of your company and get some benefits by doing so.</li>
                        </ol>
                        <p>Happy testing!</p>
                    </div>
                    <div class="modal-footer">
                      <a href="#!" class="modal-close waves-effect waves-green btn-flat">Close</a>
                    </div>
                </div>
               """

    @staticmethod
    def get_tiny_card_template(label, data):
        return """
                    <div class="col s6 m4 l3 xl2">
                      <div class="card blue-grey darken-3 absolute_cards">
                        <div class="card-content white-text">
                          <span class="card-title">{label}</span>
                          <div class="absolute_value"><b>{data}</b></div>
                        </div>
                      </div>
                    </div>
                   """.format(label=label, data=data)

    @staticmethod
    def get_table(data):
        def css():
            return """
                   <style>
                        table.dataTable thead th, table.dataTable thead td, table.dataTable.no-footer{
                        border: 0px !important;}
                        table.dataTable.no-footer{padding-bottom: 10px;}
                        #table_results tr{background: transparent;}
                        #table_results tbody tr:hover{
                            background: #1e4254;
                            cursor: pointer;
                            border-radius: 3px;}
                        #table_results_info{color: #ffffff;}
                        .dataTables_wrapper .dataTables_paginate{padding-right:50px;}
                        .dataTables_wrapper .dataTables_paginate .paginate_button.current, .dataTables_wrapper .dataTables_paginate .paginate_button {
                            color: #ffffff !important;
                            background: transparent !important;
                            border: 1px transparent!important;}
                        .dataTables_wrapper .dataTables_paginate .paginate_button.current, .dataTables_wrapper .dataTables_paginate .paginate_button.current:hover {
                            color: #ffffff !important;
                            border-bottom: 2px #fff !important;}
                        .dataTables_wrapper .dataTables_paginate .paginate_button:hover {
                            background: transparent !important;
                            border: 1px transparent!important;}
                        .dataTables_wrapper .dataTables_paginate .paginate_button:active {
                            background: transparent !important;
                            border: 1px transparent!important;}
                        .dataTables_wrapper .dataTables_paginate .paginate_button.current{
                            background: #1e4254 !important;}
                   </style>
                   """

        def js():
            return """
                   <script>
                        $(document).ready(function() {

                            var table = $("#table_results").DataTable( {
                                deferRender: true,
                                data: {data},
                                displayLength: 25,
                                order: [[ 0, "asc" ]],
                                columns: [
                                    {
                                        title: "Suite Name" ,
                                        mDataProp: "suite",
                                        defaultContent: "N/A"},
                                    {
                                        title: "Test Name" ,
                                        mDataProp: "test",
                                        defaultContent: "N/A"},
                                    {
                                        title: "Feature" ,
                                        mDataProp: "feature",
                                        defaultContent: "N/A"},
                                    {
                                        title: "Component" ,
                                        mDataProp: "component",
                                        defaultContent: "N/A"},
                                    {
                                        title: "Duration" ,
                                        mDataProp: "duration",
                                        defaultContent: "N/A"},
                                    {
                                        title: "Status" ,
                                        mDataProp: "status",
                                        defaultContent: "N/A"}],
                                createdRow: function(row, data, index){
                                    row.setAttribute("href", "#testCaseModal")
                                    row.setAttribute('class', 'modal-trigger')
                                    row.setAttribute('data-test_id', data.test_id)
                                    row.setAttribute('data-suite_id', data.suite_id)
                                }
                            });

                            var example_length = $('#table_results_length');
                            var example_filter = $('#table_results_filter');
                            example_filter.html("")
                            example_length.html("")

                             $('#table_results thead th').each(function(){
                                var title = $('#table_results thead th').eq($(this).index()).text();
                                $(this).html('<input class="deletable" id="search'+$('#table_results thead th').eq($(this).index()).text()+'" type="text" placeholder="'+title+'"/>');
                            });

                            table.columns().every(function(){
                                var tableColumn = this;
                                var input = $(this.header()).find('input');

                                input.on('change', function(){
                                    tableColumn.search(this.value).draw();
                                });
                                input.on('click', function(e){
                                    e.stopPropagation();
                                });
                                input.keyup(function(e){
                                    tableColumn.search(this.value).draw();
                                    input.focus();
                                });
                                input.keydown(function(e){
                                    if (e.keyCode == 65 && e.ctrlKey) {
                                        e.target.select()
                                    }
                                });
                            });
                        });
                   </script>
                   """.replace("{data}", str(data))

        return """
                <div class="row">
                    <div class="col s12 m12 l12 xl12">
                      <div class="card blue-grey darken-3">
                            <div class="card-content white-text">
                                <table id="table_results" class="mdl-data-table" ></table>
                                {css}
                                {js}
                            </div>
                        </div>
                    </div>
                    
                    <div id="testCaseModal" class="modal" style="background-color: #3d4f58; color: #bbc3c7">
                        <div id="testCaseModalContent" class="modal-content">
                              <div class="row"><span style="font-size: 22px;" id="test_name"></span></div>
                              <div class="row"><ul class="collapsible expandable" id="test_details"></ul></div>
                        </div>
                    </div>
                    <script>
                        $(document).ready(function(){{
                            $('#testCaseModal').modal();
                            var test_modal = $("#testCaseModal")
                            $("tr.modal-trigger").on("click", function(event){{
                                var row = $(this)[0].dataset
                                var test_id = row.test_id
                                var suite_id = row.suite_id
                                render_modal(test_id, suite_id)
                            }})

                            function render_modal(test_id, suite_id){{
                                var status = database_lol.tests[test_id]["status"]
                                var name = database_lol.tests[test_id]["name"]
                                $("#test_name").html("Test: <b>"+name+"</b>")
                                var inject_html = ""
                                var data = database_lol.tests[test_id].metrics
                                for(class_param in data){{
                                    var class_data = data[class_param]
                                    for(test_param in class_data){{
                                        var test_data = class_data[test_param]
                                        var actual_suite_param = test_data.class_param
                                        var actual_test_param = test_data.param
                                        var status = test_data.status
                                        var retry = test_data.retry
                            
                                        var params_ul = '<li class="bg">'
                            
                                        params_ul += '<div class="collapsible-header bg">'
                                            params_ul += '<div class="col xl1 badge_value"><span class="badge '+status+'">'+status+'</span></div>'
                                            params_ul += '<div class="col xl11">'
                                                params_ul += '<div data-position="bottom" data-tooltip="'+actual_suite_param+'" class="col xl6 parameter tooltipped">Class parameter: '+actual_suite_param+'</div>'
                                                params_ul += '<div data-position="bottom" data-tooltip="'+actual_test_param+'" class="col xl6 parameter tooltipped">Test parameter: '+actual_test_param+'</div>'
                                            params_ul += '</div>'
                                        params_ul += '</div>'    
                                        
                                        params_ul += '<div class="collapsible-body">'
                                        
                                        var details_ul = '<ul class="collapsible expandable">'
                                        for(index in test_data.performance){{
                                            var attempt = parseInt(index) + 1
                                             var new_li = '<li>'
                                                new_li += '<div class="collapsible-header attempt-header">'
                                                    new_li += '<div class="col xl1">'
                                                        new_li += '<i class="material-icons">fingerprint</i>'
                                                    new_li += '</div>'
                                                    new_li += '<div class="col xl2">'
                                                        new_li += '<span>Attempt: <span class="badge info-badge">'+attempt+'</span></span>'
                                                    new_li += '</div>'
                                                    new_li += '<div class="col xl3">'
                                                        new_li += '<span>Runtime: <span data-position="bottom" data-tooltip="Seconds" class="badge info-badge tooltipped">'+test_data.performance[index]+'</span></span>'
                                                    new_li += '</div>'
                                                    new_li += '<div class="col xl3">'
                                                        new_li += '<span>@beforeTest: <span class="badge info-badge">N/A</span></span>'
                                                    new_li += '</div>'
                                                    new_li += '<div class="col xl3">'
                                                        new_li += '<span>@afterTest: <span class="badge info-badge">N/A</span></span>'
                                                    new_li += '</div>'
                                                new_li += '</div>'
                                                if(test_data.tracebacks[index] != null){{
                                                    new_li += '<div class="collapsible-body"><p class="traceback">'+test_data.tracebacks[index]+'</p></div>'
                                                }}
                                             new_li += '</li>'
                                             details_ul += new_li
                                        }}
                                        details_ul += '</ul>'
                                        params_ul += details_ul
                                        params_ul += '</div>'
                                        params_ul += '</li>'
                                        inject_html += params_ul
                                    }}
                                }}
                                $("#test_details").html(inject_html)
                                $('.collapsible.expandable').collapsible();
                                $('.tooltipped').tooltip();
                            }}

                        }});
                    </script>
                </div>
               """.format(css=css(), js=js())

    @staticmethod
    def get_resource_chart_template(data):
        """
        :param data: LIST of DICTs aka
                     [{date: newDate, cpu: cpu, mem: mem},
                      {date: newDate, cpu: cpu, mem: mem}]
        :return: STRING, HTML for the CPU & MEM chart
        """
        doc_link = "<span>Resource monitoring disabled. See <a href=\"{link}\">documentation</a> to enabled it.</span>"
        data = [] if data is None else data

        def css():

            return """
                   <style>
                        #resources {
                            width: 100%;
                            height: 282px;
                        }
                   </style>
                   """

        def js():

            return """
                   <script>
                   $(document).ready(function() {
                        // Themes
                        am4core.useTheme(am4themes_dark);
                        am4core.useTheme(am4themes_animated);
                        
                        // Create chart
                        var chart = am4core.create("resources", am4charts.XYChart);
                        chart.paddingRight = 20;
                        
                        // format data
                        var data = {data};
                        for(dict in data) data[dict].date = new Date(data[dict].date)
                        
                        chart.data = data;
    
                        var dateAxis = chart.xAxes.push(new am4charts.DateAxis());
                        dateAxis.baseInterval = {
                          "timeUnit": "second",
                          "count": 1
                        };
                        dateAxis.tooltipDateFormat = "HH:mm:ss, d MMMM";
    
                        var valueAxis = chart.yAxes.push(new am4charts.ValueAxis());
                        valueAxis.tooltip.disabled = true;
                        valueAxis.title.text = "Usage %";
                        valueAxis.strictMinMax  = 100;
    
                        function createSeries(field, name){
                          var series = chart.series.push(new am4charts.LineSeries());
                          series.name = name;
                          series.dataFields.dateX = "date";
                          series.dataFields.valueY = field;
                          series.tooltipText = name+": [bold]{valueY}%[/]";
                          series.fillOpacity = 0.3;
                          return series
                        }
    
                        createSeries("cpu", "cpu")
                        createSeries("mem", "mem")
                        chart.cursor = new am4charts.XYCursor();
                        chart.cursor.lineY.opacity = 0;
                    });
                   </script>
                   """.replace("{data}", str(data))

        return """
                <div class='col s12 m12 l6 xl6'>
                  <div class='card blue-grey darken-3'>
                    <div class='card-content white-text'>
                      <span class='card-title'>CPU & Memory Usage:</span>
                      {doc_link}
                      <div id="resources"></div>
                      {css}
                      {js}
                    </div>
                  </div>
                </div>
               """.format(css=css(), js=js(), doc_link=doc_link if not data else "")

    @staticmethod
    def get_absolute_results_template(data, colors):
        """
        Colors need to come with the data as well
        :param data: LIST of DICTs aka
                     [{status: "Pass", value: 401},
                      {status: "Fail", value: 300},
                      {status: "Error", value: 200},
                      {status: "Ignore", value: 165},
                      {status: "Canceled", value: 139},
                      {status: "Skipped", value: 128}]
        :return: STRING, HTML for the Absolute Metrics chart
        """
        def css():

            return """
                   <style>
                       #absolute_metrics{
                            width: 100%;
                            height: 282px;
                       }
                   </style>
                   """

        def js():

            return """
                   <script>
                        // Theme
                        am4core.useTheme(am4themes_dark);
                        am4core.useTheme(am4themes_animated);

                        var chart = am4core.create("absolute_metrics", am4charts.PieChart);
                        chart.hiddenState.properties.opacity = 0; // this creates initial fade-in

                        chart.data = {data};
                        chart.radius = am4core.percent(70);
                        chart.innerRadius = am4core.percent(40);
                        chart.startAngle = 180;
                        chart.endAngle = 360;

                        var series = chart.series.push(new am4charts.PieSeries());
                        
                        // format colors
                        var colors = {colors}
                        for (item in colors) colors[item] = am4core.color(colors[item]) 
                        series.colors.list = colors;
                        
                        series.dataFields.value = "value";
                        series.dataFields.category = "status";
                        series.slices.template.cornerRadius = 10;
                        series.slices.template.innerCornerRadius = 7;
                        series.slices.template.draggable = true;
                        series.slices.template.inert = true;
                        series.alignLabels = false;
                        series.hiddenState.properties.startAngle = 90;
                        series.hiddenState.properties.endAngle = 90;
                   </script>
                   """.replace("{data}", str(data)).replace("{colors}", str(colors))

        return """
                <div class="col s12 m12 l6 xl6">
                  <div class="card blue-grey darken-3">
                    <div class="card-content white-text">
                      <span class="card-title">Absolute Metrics:</span>
                        <div id="absolute_metrics"></div>
                        {css}
                        {js}
                    </div>
                  </div>
                </div>
               """.format(css=css(), js=js())

    @staticmethod
    def get_stacked_bar_results_template(features_data, components_data, team_data, suites_data, tags_data):
        """
        All data must be passed in in the following format:
        [{"duration": 10,
          "measure": "Login",
          "passed": 227,
          "error": 10,
          "cancel": 12,
          "ignore": 40,
          "failed": 408
        }, {"duration": 10,
            "measure": "Dashboard",
            "passed": 371,
            "error": 10,
            "cancel": 7,
            "ignore": 38,
            "failed": 482}, ...]
        :param features_data: LIST of DICTs
        :param components_data: LIST of DICTs
        :param team_data: LIST of DICTs
        :param suites_data: LIST of DICTs
        :param tags_data: LIST of DICTs
        :return: STRING, HTML for all the charts in the tabs
        """

        series = []
        from test_junkie.reporter.reporter import Reporter
        for status in TestCategory.ALL:
            series.append({"status": status, "label": status, "color": Reporter.COLOR_MAPPING[status]})

        def css(chart_id):

            return """
                    <style>
                        #{chart_id} {{
                            width: 100%;
                            height: 400px;
                        }}
                    </style>
                   """.format(chart_id=chart_id)

        def js(chart_id, chart_data):

            return """
                   <script>
                   $(document).ready(function() {
                        // Themes begin
                        am4core.useTheme(am4themes_animated);
                        am4core.useTheme(am4themes_dark);
                        // Create chart instance
                        var chart_{chart_id} = am4core.create("{chart_id}", am4charts.XYChart);
                        
                        // Add data
                        chart_{chart_id}.data = {data}
                        
                        // Create axes
                        var categoryAxis_{chart_id} = chart_{chart_id}.xAxes.push(new am4charts.CategoryAxis());
                        categoryAxis_{chart_id}.dataFields.category = "measure";
                        categoryAxis_{chart_id}.renderer.grid.template.location = 0;
                        categoryAxis_{chart_id}.renderer.labels.template.wrap = true;
                        categoryAxis_{chart_id}.renderer.labels.template.maxWidth = 120;
                        
                        var valueAxis_{chart_id} = chart_{chart_id}.yAxes.push(new am4charts.ValueAxis());
                        valueAxis_{chart_id}.renderer.inside = true;
                        valueAxis_{chart_id}.renderer.labels.template.disabled = true;
                        valueAxis_{chart_id}.min = 0;
                        
                        var testsAxis_{chart_id} = chart_{chart_id}.yAxes.push(new am4charts.ValueAxis());
                        testsAxis_{chart_id}.title.text = "Tests";
                        testsAxis_{chart_id}.renderer.grid.template.disabled = true;
                        
                        var durationAxis_{chart_id} = chart_{chart_id}.yAxes.push(new am4charts.DurationAxis());
                        durationAxis_{chart_id}.title.text = "Average Test Duration";
                        durationAxis_{chart_id}.baseUnit = "second";
                        durationAxis_{chart_id}.renderer.grid.template.disabled = true;
                        durationAxis_{chart_id}.renderer.opposite = true;
                        durationAxis_{chart_id}.durationFormatter.durationFormat = "h'h' m'min' s'sec'";
                        
                        // Create series
                        function createSeries(field, name, color) {
                              // Set up series
                              var series_{chart_id} = chart_{chart_id}.series.push(new am4charts.ColumnSeries());
                              series_{chart_id}.name = name;
                              series_{chart_id}.yAxis = testsAxis_{chart_id};
                              series_{chart_id}.dataFields.valueY = field;
                              series_{chart_id}.dataFields.categoryX = "measure";
                              series_{chart_id}.sequencedInterpolation = true;
                            
                              // Make it stacked
                              series_{chart_id}.stacked = true;
                            
                              // Configure columns
                              series_{chart_id}.columns.template.width = am4core.percent(60);
                              series_{chart_id}.columns.template.tooltipText = "[bold]{name}[/]\\n[font-size:14px]{categoryX}: {valueY}";
                              series_{chart_id}.columns.template.stroke = am4core.color(color);
                              series_{chart_id}.columns.template.fill = am4core.color(color);
                              series_{chart_id}.columns.template.fillOpacity = 0.9;
                              return series_{chart_id};
                        }
                        
                        var required_series_{chart_id} = {series}
                        for(entry in required_series_{chart_id}){
                            createSeries(required_series_{chart_id}[entry].status, required_series_{chart_id}[entry].label, required_series_{chart_id}[entry].color)
                        }
                        
                        var durationSeries_{chart_id} = chart_{chart_id}.series.push(new am4charts.LineSeries());
                        durationSeries_{chart_id}.dataFields.valueY = "duration";
                        durationSeries_{chart_id}.dataFields.categoryX = "measure";
                        durationSeries_{chart_id}.yAxis = durationAxis_{chart_id};
                        durationSeries_{chart_id}.name = "Duration";
                        durationSeries_{chart_id}.strokeWidth = 2;
                        durationSeries_{chart_id}.stroke = am4core.color("#03a2cd");
                        durationSeries_{chart_id}.tooltip.getFillFromObject = false;
                        durationSeries_{chart_id}.tooltip.background.fill = am4core.color("#03a2cd");
                        durationSeries_{chart_id}.tooltipText = "{valueY.formatDuration()}";
                        
                        var bullet_{chart_id} = durationSeries_{chart_id}.bullets.push(new am4charts.Bullet());
                        var circle_{chart_id} = bullet_{chart_id}.createChild(am4core.Circle);
                        circle_{chart_id}.radius = 4;
                        circle_{chart_id}.fill = am4core.color("#fff");
                        circle_{chart_id}.strokeWidth = 3;
                        
                        // Legend
                        chart_{chart_id}.legend = new am4charts.Legend();
                        
                        // Add cursor
                        chart_{chart_id}.cursor = new am4charts.XYCursor();
                    });
                   </script>
                   """.replace("{data}", str(chart_data))\
                      .replace("{chart_id}", chart_id)\
                      .replace("{series}", str(series))

        def card(chart_id, title, chart_data):
            return """
                   <div id="{chart_id}_tab" class="col s12">
                       <div class="col s12 m12 l12 xl12">
                         <div class="card blue-grey darken-3">
                           <div class="card-content white-text">
                             <span class="card-title">{title}:</span>
                             <div id="{chart_id}"></div>
                             {css}
                             {js}
                           </div>
                         </div>
                       </div>
                   </div>
                   """.format(chart_id=chart_id, title=title, css=css(chart_id), js=js(chart_id, chart_data))

        features_html = card("features", "Results by Features", features_data)
        components_html = card("components", "Results by Components", components_data)
        team_html = card("owners", "Results by Assignees", team_data)
        suites_html = card("suites", "Results by Test Suites", suites_data)
        tags_html = card("tags", "Results by Tags", tags_data)

        return """
                <div class="row">
                    <div class="col s12 m12 l12 xl12">
                      <ul class="tabs transparent" style="margin-left: 20px;">
                        <li class="tab col s2"><a class="blue-text text-lighten-2 active" href="#features_tab">Features</a></li>
                        <li class="tab col s2"><a class="blue-text text-lighten-2" href="#components_tab">Components</a></li>
                        <li class="tab col s2"><a class="blue-text text-lighten-2" href="#owners_tab">Team</a></li>
                        <li class="tab col s2"><a class="blue-text text-lighten-2" href="#suites_tab">Suites</a></li>
                        <li class="tab col s2"><a class="blue-text text-lighten-2" href="#tags_tab">Tags</a></li>
                        <div class="indicator blue" style="z-index:1"></div>
                      </ul>
                    </div>
                    {features}
                    {components}
                    {team}
                    {suites}
                    {tags}
                </div>
               """.format(features=features_html, components=components_html, team=team_html,
                          suites=suites_html, tags=tags_html)

    @staticmethod
    def get_health_of_features(data):
        """
        :param data: LIST of DICTS aka
                     [{"category": "Feature A", "value": 95, "full": 100},
                      {"category": "Feature B", "value": 90, "full": 100}]
        :return: STRING, HTML for all the Health of Features chart
        """
        def css():

            return """
                    <style>
                        #health {
                            width: 100%;
                            height: 282px;
                        }
                    </style>
                   """

        def js():

            return """
                    <script>
                    $(document).ready(function() {
                        // Themes begin
                        am4core.useTheme(am4themes_dark);
                        am4core.useTheme(am4themes_animated);

                        // Create chart instance
                        var chart = am4core.create("health", am4charts.RadarChart);

                        // Add data
                        chart.data = {data};

                        // Make chart not full circle
                        chart.startAngle = -90;
                        chart.endAngle = 180;
                        chart.innerRadius = am4core.percent(20);

                        // Set number format
                        chart.numberFormatter.numberFormat = "#.#'%'";

                        // Create axes
                        var categoryAxis = chart.yAxes.push(new am4charts.CategoryAxis());
                        categoryAxis.dataFields.category = "category";
                        categoryAxis.renderer.grid.template.location = 0;
                        categoryAxis.renderer.grid.template.strokeOpacity = 0;
                        categoryAxis.renderer.labels.template.horizontalCenter = "right";
                        categoryAxis.renderer.labels.template.fontWeight = 500;
                        categoryAxis.renderer.labels.template.adapter.add("fill", function(fill, target) {
                          return (target.dataItem.index >= 0) ? chart.colors.getIndex(target.dataItem.index) : fill;
                        });
                        categoryAxis.renderer.minGridDistance = 10;

                        var valueAxis = chart.xAxes.push(new am4charts.ValueAxis());
                        valueAxis.renderer.grid.template.strokeOpacity = 0;
                        valueAxis.min = 0;
                        valueAxis.max = 100;
                        valueAxis.strictMinMax = true;

                        // Create series
                        var series1 = chart.series.push(new am4charts.RadarColumnSeries());
                        series1.dataFields.valueX = "full";
                        series1.dataFields.categoryY = "category";

                        series1.columns.template.fill = new am4core.InterfaceColorSet().getFor("alternativeBackground");
                        series1.columns.template.fillOpacity = 0.08;
                        series1.columns.template.cornerRadiusTopLeft = 20;
                        series1.columns.template.strokeWidth = 0;
                        series1.columns.template.radarColumn.cornerRadius = 20;

                        var series2 = chart.series.push(new am4charts.RadarColumnSeries());
                        series2.dataFields.valueX = "value";
                        series2.dataFields.categoryY = "category";
                        series2.clustered = false;
                        series2.columns.template.strokeWidth = 0;
                        series2.columns.template.tooltipText = "{category}: [bold]{value}%[/]";
                        series2.columns.template.radarColumn.cornerRadius = 20;

                        series2.columns.template.adapter.add("fill", function(fill, target) {
                          return chart.colors.getIndex(target.dataItem.index);
                        });

                        // Add cursor
                        chart.cursor = new am4charts.RadarCursor();
                    });
                    </script>
                   """.replace("{data}", str(data))

        return """
                <div class='col s12 m12 l6 xl6'>
                  <div class='card blue-grey darken-3'>
                    <div class='card-content white-text'>
                      <span class='card-title'>Health of Features:</span>
                      <div id="health"></div>
                      {css}
                      {js}
                    </div>
                  </div>
                </div>
               """.format(css=css(), js=js())

    @staticmethod
    def get_suggestions(data):

        def css():

            return """
                    <style>
                        #suggestions {
                            width: 100%;
                            height: 282px;
                        }
                    </style>
                   """

        def js():

            return """
                   <script>
                   </script>
                   """

        return """
                <div class="col s12 m12 l6 xl6">
                  <div class="card blue-grey darken-3">
                    <div class="card-content white-text">
                      <span class="card-title">Suggestions:</span>
                      <div id="suggestions">This feature is still in development</div>
                      {css}
                      {js}
                    </div>
                  </div>
                </div>
               """.format(css=css(), js=js())



class TestCategory:

    SUCCESS = "success"
    FAIL = "fail"
    IGNORE = "ignore"
    ERROR = "error"
    SKIP = "skip"
    CANCEL = "cancel"
    ALL = [SUCCESS, FAIL, IGNORE, ERROR, SKIP, CANCEL]
    ALL_UN_SUCCESSFUL = [FAIL, IGNORE, ERROR]


class SuiteCategory:

    SUCCESS = "success"
    FAIL = "fail"
    SKIP = "skip"
    CANCEL = "cancel"
    IGNORE = "ignore"
    ALL = [SUCCESS, FAIL, SKIP, CANCEL, IGNORE]
    ALL_UN_SUCCESSFUL = [FAIL, IGNORE]

    """
    This status means something failed outside of Test Junkie,
    like custom event listener function was missing required arguments in the signature.
    This status is used as fallback and is not set by Test Junkie during test execution.
    Tests that error out, suite should be marked with the FAIL status.
    Thus, this status is not part of the ALL nor ALL_UN_SUCCESSFUL list and should stay that way
    """
    ERROR = "error"


class DecoratorType:

    TEST_SUITE = "testSuite"
    BEFORE_CLASS = "beforeClass"
    BEFORE_TEST = "beforeTest"
    AFTER_TEST = "afterTest"
    AFTER_CLASS = "afterClass"
    TEST_CASE = "testCase"


class Event:

    ON_SUCCESS = 1
    ON_FAILURE = 2
    ON_ERROR = 3
    ON_SKIP = 4
    ON_IGNORE = 5
    ON_CANCEL = 6
    ON_CLASS_SKIP = 7
    ON_CLASS_CANCEL = 8
    ON_BEFORE_CLASS_ERROR = 9
    ON_BEFORE_CLASS_FAIL = 10
    ON_AFTER_CLASS_ERROR = 11
    ON_AFTER_CLASS_FAIL = 12
    ON_CLASS_IN_PROGRESS = 13
    ON_CLASS_COMPLETE = 14
    ON_CLASS_IGNORE = 15


class DocumentationLinks:

    SUITE_DECORATOR = "https://github.com/ArturSpirin/test_junkie#suite"
    TEST_DECORATOR = "https://github.com/ArturSpirin/test_junkie#test"
    ON_CLASS_IGNORE = "https://github.com/ArturSpirin/test_junkie#on-class-ignore"
    ON_TEST_IGNORE = "https://github.com/ArturSpirin/test_junkie#on-ignore"
    PARAMETERIZED_TESTS = "https://github.com/ArturSpirin/test_junkie#parameterized-tests"
    LISTENERS = "https://github.com/ArturSpirin/test_junkie#test-listeners"
    TAGS = "https://github.com/ArturSpirin/test_junkie#tags"
    SPONSOR_PATREON = "https://www.patreon.com/join/arturspirin"
    RUNNER_OBJECT = "https://github.com/ArturSpirin/test_junkie#runner-object"



import time
from test_junkie.decorators import Suite, test, beforeTest, beforeClass
from test_junkie.meta import Meta
from tests.junkie_suites.Constants import Constants


def test_func():
    print("evaluating...")
    time.sleep(2)
    return [1, 2]


@Suite(retry=2, feature="Login", owner="Mike", parameters=["1122222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222", 22])
class LoginSuite:

    @beforeClass()
    def before_class(self):
        raise Exception("Find me!")

    @beforeTest()
    def before_test(self):
        print("Winning!")

    @test(component="Login Page", skip_before_test=True,
          tags=["positive_flow", "ui", "auth"])
    def positive_login(self):
        time.sleep(1)

    @test(priority=1, skip_before_test=True,
          parallelized_parameters=True,
          component="Login Page",
          tags=["negative_flow", "ui"],
          parameters=[1, 2, 3])
    def negative_login(self, parameter, suite_parameter):
        # time.sleep(3.3)
        if parameter == 1:
            return
        Constants.raise_large_exception()
        # raise AssertionError("Missing error message on negative login attempt: {}".format(parameter))

    @test(component="Session", owner="Victor", skip_before_test=True,
          tags=["positive_flow", "ui", "auth", "session"])
    def session_timeout_after_2h(self):
        pass

    @test(component="Session", owner="Victor",
          tags=["positive_flow", "ui", "auth", "session"])
    def session_timeout_after_1h(self):
        pass

    @test(component="Session", owner="Victor",
          tags=["negative_flow", "ui", "auth", "session"])
    def logout_after_session_timeout(self):
        pass

    @test(component="Login Page", skip_before_test=True,
          tags=["negative_flow", "ui"])
    def negative_login_attempt_limit(self):
        # time.sleep(7.3)
        pass



from test_junkie.decorators import beforeClass, afterClass, afterTest, beforeTest


class TestRules:

    @beforeClass()
    def before_class(self):
        pass

    @beforeTest()
    def before_test(self):
        # write your code here
        pass

    @afterTest()
    def after_test(self):
        # write your code here
        pass

    @afterClass()
    def after_class(self):
        # write your code here
        pass



import time
from test_junkie.decorators import Suite, test
from tests.junkie_suites.TestListener import TestListener


@Suite(retry=2,
       listener=TestListener,
       meta={"name": "Suite B", "known_bugs": []},
       parameters=[1, 2], priority=1, feature="Store",
       owner="George")
class ShoppingCartSuite:

    @test(priority=1,
          component="Shopping Cart",
          tags=["ui", "cart", "positive_flow"])
    def add_to_cart(self):
        time.sleep(1)
        print("Finished SUITE B / TEST A")

    @test(priority=1,
          component="Shopping Cart",
          tags=["ui", "cart", "positive_flow"])
    def remove_from_cart(self):
        time.sleep(2)
        pass

    @test(priority=1,
          component="Shopping Cart",
          tags=["ui", "cart", "positive_flow"])
    def increase_quantity(self):
        time.sleep(1.2)
        pass

    @test(priority=1,
          component="Shopping Cart",
          tags=["ui", "cart", "positive_flow"])
    def decrease_quantity(self):
        time.sleep(1.2)
        pass

    @test(priority=1,
          component="Shopping Cart",
          tags=["ui", "cart", "positive_flow"])
    def save_for_later(self):
        time.sleep(1.2)
        pass



from test_junkie.rules import Rules


class TestRules(Rules):

    def __init__(self, **kwargs):

        Rules.__init__(self, **kwargs)

    def before_class(self):
        # write your code here
        pass

    def before_test(self):
        # write your code here
        pass

    def after_test(self):
        # write your code here
        pass

    def after_class(self):
        # write your code here
        pass



import pprint

from test_junkie.errors import BadParameters, BadSignature
from test_junkie.runner import Runner
from tests.QualityManager import QualityManager
from tests.junkie_suites.IgnoreSuite import IgnoreSuiteBoundMethod, IgnoreSuiteFunction, IgnoreSuiteClassic, \
    IgnoreSuiteClassic2, IgnoreSuiteClassic3

runner = Runner(IgnoreSuiteBoundMethod)
runner.run()
results = runner.get_executed_suites()

pprint.pprint(results[0].metrics.get_metrics())
for test in results[0].get_test_objects():
    print(test.get_function_name())
    pprint.pprint(test.metrics.get_metrics())


def test_class_metrics():

    metrics = results[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics,
                                       expected_status="ignore",
                                       expected_retry_count=0)


def test_test_metrics():

    assert results[0].get_test_objects()
    for test in results[0].get_test_objects():

        properties = test.metrics.get_metrics()
        assert len(properties) == 0


runner = Runner(IgnoreSuiteFunction)
runner.run()
results = runner.get_executed_suites()

pprint.pprint(results[0].metrics.get_metrics())
for test in results[0].get_test_objects():
    print(test.get_function_name())
    pprint.pprint(test.metrics.get_metrics())


def test_class_metrics2():

    metrics = results[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics,
                                       expected_status="ignore",
                                       expected_retry_count=0)


def test_test_metrics2():

    assert results[0].get_test_objects()
    for test in results[0].get_test_objects():

        properties = test.metrics.get_metrics()
        assert len(properties) == 0


runner = Runner(IgnoreSuiteClassic)
runner.run()
results = runner.get_executed_suites()

pprint.pprint(results[0].metrics.get_metrics())
for test in results[0].get_test_objects():
    print(test.get_function_name())
    pprint.pprint(test.metrics.get_metrics())


def test_class_metrics3():

    metrics = results[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics,
                                       expected_status="ignore",
                                       expected_retry_count=0)


def test_test_metrics3():

    assert results[0].get_test_objects()
    for test in results[0].get_test_objects():

        properties = test.metrics.get_metrics()
        assert len(properties) == 0


runner = Runner(IgnoreSuiteClassic2)
runner.run()
results = runner.get_executed_suites()

pprint.pprint(results[0].metrics.get_metrics())
for test in results[0].get_test_objects():
    print(test.get_function_name())
    pprint.pprint(test.metrics.get_metrics())


def test_class_metrics4():

    metrics = results[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics,
                                       expected_status="ignore",
                                       expected_retry_count=0)


def test_test_metrics4():

    assert results[0].get_test_objects()
    for test in results[0].get_test_objects():

        properties = test.metrics.get_metrics()
        assert len(properties) == 0


runner = Runner(IgnoreSuiteClassic3)
runner.run()
results = runner.get_executed_suites()

pprint.pprint(results[0].metrics.get_metrics())
for test in results[0].get_test_objects():
    print(test.get_function_name())
    pprint.pprint(test.metrics.get_metrics())


def test_class_metrics5():

    metrics = results[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics,
                                       expected_status="ignore",
                                       expected_retry_count=0)


def test_test_metrics5():

    assert results[0].get_test_objects()
    for test in results[0].get_test_objects():

        properties = test.metrics.get_metrics()
        assert len(properties) == 0


def test_wrong_params():

    try:
        from tests.junkie_suites.error_handling import ErrorSuite1
        raise AssertionError("This test must raise exception, as wrang datatype is used for parameters")
    except Exception as error:
        assert isinstance(error, BadParameters), "Exception must be raised for bad parameters"


def test_wrong_params2():

    try:
        from tests.junkie_suites.error_handling import ErrorSuite2
        raise AssertionError("This test must raise exception, as wrang datatype is used for parameters")
    except Exception as error:
        assert isinstance(error, BadSignature), "Exception must be raised for bad signature"


def test_wrong_params3():

    try:
        from tests.junkie_suites.error_handling.ErrorSuite3 import ErrorSuite3
        raise AssertionError("This test must raise exception, as wrang datatype is used for parameters")
    except Exception as error:
        assert isinstance(error, BadParameters), "Exception must be raised for bad parameters"



import pprint
import sys
# from tests.junkie_suites.IgnoreSuite import IgnoreSuiteWrongDatatype
from playground.suiteB import ShoppingCartSuite
from playground.suiteC import AuthApiSuite
from playground.suiteD import NewProductsSuite
from test_junkie.objects import Limiter

sys.path.insert(1, __file__.split("playground")[0])
from test_junkie.reporter.reporter import Reporter
from test_junkie.debugger import LogJunkie
LogJunkie.enable_logging(10)

from playground.suiteA import LoginSuite
from test_junkie.runner import Runner
Limiter.ACTIVE = True
runner = Runner([LoginSuite], monitor_resources=True,
                html_report="E:\Development\\test_junkie\\playground\\report.html")

aggregator = runner.run(test_multithreading_limit=10)
suite_objects = runner.get_executed_suites()

for suite in suite_objects:
    print(suite.get_class_name())
    tests = suite.get_test_objects()
    print(suite.metrics.get_metrics())
    print(suite.get_data_by_tags())
    for test in tests:
        print(test.metrics.get_metrics())

# tags = aggregator.get_report_by_tags()
# features = aggregator.get_report_by_features()
# totals = aggregator.get_basic_report()
# owners = aggregator.get_report_by_owner()
# reporter = Reporter("E:\Development\\test_junkie\\test_junkie\.resources", features, totals)
# pprint.pprint(reporter.get_dataset_per_feature())
# pprint.pprint(reporter.get_absolute_results())
# pprint.pprint(aggregator.get_report_by_owner())



import inspect
import threading
import time
import traceback
import sys
from test_junkie.constants import SuiteCategory, TestCategory, Event, DocumentationLinks
from test_junkie.debugger import LogJunkie
from test_junkie.decorators import DecoratorType, synchronized
from test_junkie.errors import ConfigError, TestJunkieExecutionError, TestListenerError, BadParameters, BadSignature
from test_junkie.listener import Listener
from test_junkie.metrics import Aggregator, ResourceMonitor
from test_junkie.parallels import ParallelProcessor
from test_junkie.builder import Builder
from test_junkie.reporter.reporter import Reporter


class Runner:

    __STATS = {}

    def __init__(self, suites, **kwargs):

        self.__stats = {}

        self.__suites = suites
        if not isinstance(self.__suites, list):
            self.__suites = [self.__suites]

        self.__suites = self.__prioritize(suites=self.__suites)
        for suite in self.__suites:
            suite_object = Builder.get_execution_roster().get(suite, None)
            suite_object.update_test_objects(self.__prioritize(suite_object=suite_object))
            Runner.__process_owners(suite_object)

        self.__kwargs = kwargs

        self.__processor = None
        self.__run_config = {}

        self.__cancel = False

        self.__executed_suites = []
        self.__active_suites = []

    def __monitoring_enabled(self):

        return self.__kwargs.get("monitor_resources", False)

    def __create_xml_report(self):

        def __update_tag_stats(tag, status):

            tag.set("tests", str(int(suite.get("tests")) + 1))
            if status == TestCategory.SUCCESS:
                tag.set("passed", str(int(tag.get("passed")) + 1))
            else:
                tag.set("failures", str(int(tag.get("failures")) + 1))
            return tag

        if self.__kwargs.get("xml_report", None) is not None:
            write_file = self.__kwargs.get("xml_report")
            import os
            from xml.etree.ElementTree import ElementTree, Element, SubElement
            import xml
            if not os.path.exists(write_file):
                request = Element("root")
                ElementTree(request).write(write_file)

            xml_file = xml.etree.ElementTree.parse(write_file)
            root = xml_file.getroot()

            suites = self.get_executed_suites()

            for suite_object in suites:

                test_suite = suite_object.get_class_name()
                tests = suite_object.get_test_objects()

                for test_object in tests:

                    test_name = test_object.get_function_name()
                    test_metrics = test_object.metrics.get_metrics()

                    for class_param, class_param_data in test_metrics.items():
                        for param, param_data in class_param_data.items():

                            test_status = param_data["status"]
                            if test_status != TestCategory.SUCCESS:
                                test_status = "failure"
                            suite_found = False

                            for suite in root.iter("testsuite"):
                                suite_found = suite.attrib["name"] == test_suite
                                if suite_found:
                                    __update_tag_stats(suite, test_status)
                                    test = Element("testcase", name=str(test_name), status=str(test_status))
                                    if test_status == "failure":
                                        failure = Element("failure", type="failure")
                                        test.append(failure)
                                    suite.append(test)
                                    ElementTree(root).write(write_file)
                                    break

                            if not suite_found:
                                suite = SubElement(root, "testsuite", name=test_suite,
                                                   tests="0", passed="0", failures="0")
                                __update_tag_stats(suite, test_status)
                                test = SubElement(suite, "testcase", name=str(test_name), status=str(test_status))
                                if test_status == "failure":
                                    SubElement(test, "failure", type="failure")
                                ElementTree(root).write(write_file)

    def __create_html_report(self, reporter):

        if self.__kwargs.get("html_report", None) is not None:
            reporter.generate_html_report(self.__kwargs.get("html_report"))

    @staticmethod
    def __process_owners(suite_object):

        tests = suite_object.get_test_objects()
        for test in tests:
            if test.get_owner() is None:
                test.get_kwargs().update({"owner": suite_object.get_owner()})

    @staticmethod
    def __prioritize(suites=None, suite_object=None):
        """
        This function will order the lists of suites and tests according to the priority set by user
        - Suites/Tests with priorities, will be pushed towards the front according to the priorities defined
        - Suites/Tests with no priority defined, will be pushed towards the middle
        - Suites/Tests with no priority defined and not parallelized, will be pushed towards the end
        :param suites: LIST of class objects decorated with @Suite
        :param suite_object: SuiteObject
        :return: LIST ordered list of suites or tests
        """
        ordered = []
        priorities = {}
        no_priority = []
        not_parallelized = []

        items = suites if suites is not None else suite_object.get_test_objects()

        for item in items:

            if suites is not None:
                suite_object = Builder.get_execution_roster().get(item, None)
                priority = suite_object.get_priority()
                is_parallelized = suite_object.is_parallelized()
            else:
                priority = item.get_priority()
                is_parallelized = item.is_parallelized()
            if priority is None:
                if is_parallelized:
                    no_priority.append(item)
                else:
                    not_parallelized.append(item)
            else:
                if priority not in priorities:
                    priorities.update({priority: [item]})
                else:
                    priorities[priority].append(item)

        ordered_priorities = list(priorities.keys())
        ordered_priorities.sort()
        for priority in ordered_priorities:
            for item in priorities[priority]:
                ordered.append(item)

        ordered += no_priority + not_parallelized
        return ordered

    def get_executed_suites(self):
        """
        :return: LIST of SuiteObjects that begun executing
        """
        return self.__executed_suites

    def cancel(self):
        """
        Flips the switch to cancel the execution of tests
        :return: None
        """
        self.__cancel = True

    def run(self, **kwargs):
        """
        Initiates the execution process that runs tests
        :return: None
        """
        initial_start_time = time.time()
        resource_monitor = None
        try:
            if self.__monitoring_enabled():
                resource_monitor = ResourceMonitor()
                resource_monitor.start()
            self.__run_config = kwargs
            self.__processor = ParallelProcessor(**kwargs)

            parallels = []
            while self.__suites:
                for suite in list(self.__suites):
                    suite_object = Builder.get_execution_roster().get(suite, None)
                    if suite_object is not None:
                        if self.__processor.suite_multithreading() and suite_object.is_parallelized():
                            while True:
                                if self.__processor.suite_qualifies(suite_object):
                                    self.__executed_suites.append(suite_object)
                                    parallels.append(ParallelProcessor
                                                     .run_suite_in_a_thread(self.__run_suite, suite_object))
                                    self.__suites.remove(suite)
                                    break
                                elif suite_object.get_priority() is None:
                                    break
                                else:
                                    time.sleep(1)
                        else:
                            if not suite_object.is_parallelized():
                                LogJunkie.debug("Cant run suite: {} in parallel with any other suites. Waiting for "
                                                "parallel suites to finish so I can run it by itself."
                                                .format(suite_object.get_class_object()))
                                ParallelProcessor.wait_for_parallels_to_finish(parallels)
                            self.__executed_suites.append(suite_object)
                            self.__run_suite(suite_object)
                            self.__suites.remove(suite)
                    else:
                        LogJunkie.warn("Suite: {} not found! Make sure that your input is correct. "
                                       "If it is, make sure the use of Test Junkie's decorators "
                                       "is correct.".format(suite))
                        self.__suites.remove(suite)
                LogJunkie.debug("{} Suite(s) left in queue.".format(len(self.__suites)))
                time.sleep(1)

            for parallel in parallels:
                parallel.join()
        finally:
            if self.__monitoring_enabled():
                resource_monitor.shutdown()

        runtime = time.time() - initial_start_time
        print("========== Test Junkie finished in {:0.2f} seconds ==========".format(runtime))
        aggregator = Aggregator(self.get_executed_suites())
        Aggregator.present_console_output(aggregator)
        if self.__kwargs.get("html_report", False):
            reporter = Reporter(monitoring_file=resource_monitor.get_file_path()
                                if resource_monitor is not None else None,
                                runtime=runtime,
                                aggregator=aggregator)
            self.__create_html_report(reporter)
        self.__create_xml_report()
        return aggregator

    @staticmethod
    def __validate_suite_parameters(suite):
        parameters = suite.get_parameters(process_functions=True)
        if not parameters or not isinstance(parameters, list):
            if isinstance(parameters, list):
                return BadParameters("Argument: \"parameters\" in @Suite() decorator returned empty: <class 'list'>. "
                                     "For more info, see this explanation: {}"
                                     .format(DocumentationLinks.ON_CLASS_IGNORE))
            else:
                return BadParameters("Argument: \"parameters\" in @Suite() decorator must be of type: <class 'list'> "
                                     "but found: {}. For more info, see @Suite() decorator documentation: {}"
                                     .format(type(parameters), DocumentationLinks.SUITE_DECORATOR))
        return False

    @staticmethod
    def __validate_test_parameters(test):
        parameters = test.get_parameters(process_functions=True)
        if not parameters or not isinstance(parameters, list):
            if isinstance(parameters, list):
                return BadParameters("Argument: \"parameters\" in @test() decorator returned empty: <class 'list'>. "
                                     "For more info, see: {}".format(DocumentationLinks.ON_TEST_IGNORE))
            else:
                return BadParameters("Argument: \"parameters\" in @test() decorator must be of type: <class 'list'> "
                                     "but found: {}. For more info, see @test() decorator documentation: {}"
                                     .format(type(parameters), DocumentationLinks.TEST_DECORATOR))

    def __run_suite(self, suite):
        suite_start_time = time.time()
        unsuccessful_tests = None
        bad_params = Runner.__validate_suite_parameters(suite)
        if not suite.can_skip(self.__run_config.get("features", None)) and not self.__cancel and not bad_params:
            Runner.__process_event(event=Event.ON_CLASS_IN_PROGRESS, suite=suite)
            for suite_retry_attempt in range(1, suite.get_retry_limit() + 1):
                if suite_retry_attempt == 1 or suite.get_status() in SuiteCategory.ALL_UN_SUCCESSFUL:

                    for class_param in suite.get_parameters(process_functions=True):

                        LogJunkie.debug("Running suite: {}".format(suite.get_class_object()))
                        LogJunkie.debug("Suite Retry {}/{} with Param: {}"
                                        .format(suite_retry_attempt, suite.get_retry_limit(), class_param))

                        before_class_error = Runner.__run_before_class(suite, class_param)

                        if suite_retry_attempt > 1:
                            unsuccessful_tests = suite.get_unsuccessful_tests()
                            LogJunkie.debug("There are {} unsuccessful tests that need to be retried"
                                            .format(len(unsuccessful_tests)))
                            if not unsuccessful_tests:
                                break
                        parallels = []

                        tests = list(suite.get_test_objects())
                        while tests:
                            for test in list(tests):

                                test_start_time = time.time()  # will use in case of a failure in context of this loop

                                if not Runner.__positive_skip_condition(test=test, run_config=self.__run_config) and \
                                        Runner.__runnable_tags(test=test, run_config=self.__run_config):

                                    if not test.is_parallelized():
                                        LogJunkie.debug("Cant run test: {} in parallel with any other tests"
                                                        .format(test.get_function_object()))
                                        ParallelProcessor.wait_for_parallels_to_finish(parallels)
                                    while self.__processor.test_limit_reached(parallels):
                                        time.sleep(1)
                                    bad_params = Runner.__validate_test_parameters(test)
                                    if bad_params is not None:
                                        tests.remove(test)
                                        test.metrics.update_metrics(status=TestCategory.IGNORE,
                                                                    start_time=test_start_time, exception=bad_params)
                                        Runner.__process_event(event=Event.ON_IGNORE, suite=suite, test=test,
                                                               class_param=class_param, error=bad_params)
                                        continue
                                    for param in test.get_parameters(process_functions=True):
                                        if unsuccessful_tests is not None:
                                            if not test.is_qualified_for_retry(param, class_param=class_param):
                                                # If does not qualify with current parameter, will move to the next
                                                continue
                                        if ((self.__processor.test_multithreading()
                                             and param is None) or (self.__processor.test_multithreading()
                                                                    and test.parallelized_parameters()
                                                                    and param is not None)):
                                            while True:
                                                if self.__processor.test_qualifies(suite, test):
                                                    while self.__processor.test_limit_reached(parallels):
                                                        time.sleep(1)
                                                    parallels.append(
                                                        self.__processor.run_test_in_a_thread(Runner.__run_test,
                                                                                              suite, test, param,
                                                                                              class_param,
                                                                                              before_class_error,
                                                                                              self.__cancel))
                                                    break
                                                elif test.get_priority() is None:
                                                    break
                                                else:
                                                    time.sleep(1)
                                        else:
                                            Runner.__run_test(suite=suite, test=test,
                                                              parameter=param,
                                                              class_parameter=class_param,
                                                              before_class_error=before_class_error,
                                                              cancel=self.__cancel)
                                    tests.remove(test)

                                else:
                                    tests.remove(test)
                                    test.metrics.update_metrics(status=TestCategory.SKIP, start_time=test_start_time)
                                    Runner.__process_event(event=Event.ON_SKIP, suite=suite, test=test,
                                                           class_param=class_param)

                        ParallelProcessor.wait_for_parallels_to_finish(parallels)
                        Runner.__run_after_class(suite, class_param)
                    suite.metrics.update_suite_metrics(status=SuiteCategory.FAIL
                                                       if suite.has_unsuccessful_tests() else SuiteCategory.SUCCESS,
                                                       start_time=suite_start_time)
            Runner.__process_event(event=Event.ON_CLASS_COMPLETE, suite=suite)
        elif self.__cancel:
            suite.metrics.update_suite_metrics(status=SuiteCategory.CANCEL, start_time=suite_start_time)
            Runner.__process_event(event=Event.ON_CLASS_CANCEL, suite=suite)
        elif bad_params:
            suite.metrics.update_suite_metrics(status=SuiteCategory.IGNORE, start_time=suite_start_time,
                                               initiation_error=bad_params)
            Runner.__process_event(event=Event.ON_CLASS_IGNORE, suite=suite)
        else:
            suite.metrics.update_suite_metrics(status=SuiteCategory.SKIP, start_time=suite_start_time)
            Runner.__process_event(event=Event.ON_CLASS_SKIP, suite=suite)

    @staticmethod
    def __run_test(suite, test, parameter=None, class_parameter=None, before_class_error=None, cancel=False):

        def run_before_test():
            try:
                if not test.skip_before_test_rule():
                    suite.get_rules().before_test()
                if not test.skip_before_test():
                    before_test_error = Runner.__process_decorator(decorator_type=DecoratorType.BEFORE_TEST,
                                                                   suite=suite, class_parameter=class_parameter,
                                                                   test=test, parameter=parameter)
                    if before_test_error is not None:
                        process_failure(before_test_error, pre_processed=True)
                        return False
                return True
            except Exception as before_test_error:
                process_failure(before_test_error)
                return False

        def process_failure(error, pre_processed=False):
            if not isinstance(error, TestJunkieExecutionError):
                if pre_processed:
                    trace = error.message if sys.version_info[0] < 3 else str(error)
                else:
                    trace = traceback.format_exc()
                __category, __event = TestCategory.ERROR, Event.ON_ERROR
                if isinstance(error, AssertionError):
                    __category, __event = TestCategory.FAIL, Event.ON_FAILURE
                test.metrics.update_metrics(status=__category,
                                            start_time=test_start_time,
                                            param=parameter,
                                            class_param=class_parameter,
                                            exception=error,
                                            formatted_traceback=trace)
                Runner.__process_event(event=__event, suite=suite, test=test, error=error,
                                       class_param=class_parameter, param=parameter, formatted_traceback=trace)
            else:
                raise error

        def run_after_test(_record_test_failure=True):
            try:
                # Running after test functions
                if not test.skip_after_test():
                    after_test_error = Runner.__process_decorator(decorator_type=DecoratorType.AFTER_TEST,
                                                                  suite=suite, class_parameter=class_parameter,
                                                                  test=test, parameter=parameter)
                    if after_test_error is not None:
                        process_failure(after_test_error, pre_processed=True)
                        return False
                if not test.skip_after_test_rule():
                    suite.get_rules().after_test()
                return True
            except Exception as after_test_error:
                if _record_test_failure:
                    process_failure(after_test_error)
                return False

        test_start_time = time.time()
        if before_class_error is not None or cancel:
            _status = TestCategory.IGNORE if not cancel else TestCategory.CANCEL
            _event = Event.ON_IGNORE if not cancel else Event.ON_CANCEL
            test.metrics.update_metrics(status=_status,
                                        start_time=test_start_time,
                                        param=parameter,
                                        class_param=class_parameter,
                                        exception=before_class_error["exception"],
                                        formatted_traceback=before_class_error["traceback"])
            Runner.__process_event(event=_event, error=before_class_error["exception"], suite=suite, test=test,
                                   class_param=class_parameter, param=parameter,
                                   formatted_traceback=before_class_error["traceback"])
            return

        if not test.accepts_suite_parameters():
            # for reporting purposes, so reports are properly nested
            class_parameter = None

        status = test.get_status(parameter, class_parameter)
        if not test.accepts_suite_parameters() and status is not None:
            """
            making sure that we do not run tests with suite parameters if suite parameters are not accepted in the
            test signature. But we still want to run all other parameters and tests without any params.
            Also we want to make sure we honor the suite level retries
            """
            if status != TestCategory.SUCCESS:  # test already ran and was unsuccessful
                # if it did not reach its max retry limit, will rerun it again
                max_retry_allowed = suite.get_retry_limit() * test.get_retry_limit()
                actual_retries = test.get_number_of_actual_retries(parameter, class_parameter)
                if actual_retries >= max_retry_allowed:
                    return
            else:
                return

        for retry_attempt in range(1, test.get_retry_limit() + 1):
            if test.is_qualified_for_retry(parameter, class_param=class_parameter):
                LogJunkie.debug("\n===============Running test==================\n"
                                "Test Case: {}\n"
                                "Test Suite: {}\n"
                                "Test Parameter: {}\n"
                                "Class Parameter: {}\n"
                                "Retry Attempt: {}/{}\n"
                                "============================================="
                                .format(test.get_function_name(), suite.get_class_name(), parameter, class_parameter,
                                        retry_attempt, test.get_retry_limit()))
                record_test_failure = True
                try:
                    if run_before_test() is False:  # if before test failed, moving on without running the test
                        continue  # everything recorded at this point in the metrics and flow is solid
                    # Running actual test
                    start_time = time.time()
                    Runner.__process_decorator(decorator_type=DecoratorType.TEST_CASE, suite=suite,
                                               test=test, parameter=parameter, class_parameter=class_parameter)
                    runtime = time.time() - start_time
                except Exception as test_error:
                    process_failure(test_error)
                    record_test_failure = False  # already recorded the failure just above this

                if run_after_test(record_test_failure) is True:  # if did not fail, test is OK
                    if record_test_failure:  # Test failed and failure was already recorded thus can't pass it
                        test.metrics.update_metrics(status=TestCategory.SUCCESS, start_time=None, param=parameter,
                                                    class_param=class_parameter, runtime=runtime)
                        Runner.__process_event(event=Event.ON_SUCCESS, suite=suite, test=test,
                                               class_param=class_parameter, param=parameter)
                        return

    @staticmethod
    def __run_before_class(suite, class_parameter=None):
        try:
            suite.get_rules().before_class()
            before_class_error = Runner.__process_decorator(decorator_type=DecoratorType.BEFORE_CLASS,
                                                            suite=suite, class_parameter=class_parameter)
            if before_class_error is not None:
                raise before_class_error
            return
        except Exception as error:
            trace = traceback.format_exc()
            if isinstance(error, AssertionError):
                Runner.__process_event(event=Event.ON_BEFORE_CLASS_FAIL, suite=suite, error=error,
                                       class_param=class_parameter, formatted_traceback=trace)
            else:
                Runner.__process_event(event=Event.ON_BEFORE_CLASS_ERROR, suite=suite, error=error,
                                       class_param=class_parameter, formatted_traceback=trace)
            return {"exception": error, "traceback": trace}

    @staticmethod
    def __run_after_class(suite, class_parameter=None):
        try:
            after_class_error = Runner.__process_decorator(decorator_type=DecoratorType.AFTER_CLASS,
                                                           suite=suite, class_parameter=class_parameter)
            if after_class_error is not None:
                raise after_class_error
            suite.get_rules().after_class()
        except Exception as error:
            trace = traceback.format_exc()
            if isinstance(error, AssertionError):
                Runner.__process_event(event=Event.ON_AFTER_CLASS_FAIL, suite=suite, error=error,
                                       class_param=class_parameter, formatted_traceback=trace)
            else:
                Runner.__process_event(event=Event.ON_AFTER_CLASS_ERROR, suite=suite, error=error,
                                       class_param=class_parameter, formatted_traceback=trace)

    @staticmethod
    # @synchronized(threading.Lock())
    def __process_decorator(decorator_type, suite, test=None, parameter=None,  class_parameter=None):

        def update_metrics(error=None, _trace=None):
            suite.metrics.update_decorator_metrics(decorator_type, start_time, error, _trace)
            if decorator_type in [DecoratorType.BEFORE_TEST, DecoratorType.AFTER_TEST]:
                test.metrics.update_metrics(status=None, start_time=start_time, param=parameter,
                                            class_param=class_parameter, decorator=decorator_type,
                                            formatted_traceback=_trace)
        start_time = time.time()
        if DecoratorType.TEST_CASE != decorator_type:
            functions_list = suite.get_decorated_definition(decorator_type)
            for func in functions_list:
                try:
                    # deprecated but supports Python 2
                    if "suite_parameter" in inspect.getargspec(func["decorated_function"]).args:
                        func["decorated_function"](suite.get_class_object()(), suite_parameter=class_parameter)
                    else:
                        func["decorated_function"](suite.get_class_object()())
                except Exception as decorator_error:
                    trace = traceback.format_exc()
                    update_metrics(decorator_error, trace)
                    return AssertionError(trace) \
                        if isinstance(decorator_error, AssertionError) else Exception(trace)
                update_metrics()
        else:
            if test.accepts_test_and_suite_parameters():
                test.get_function_object()(suite.get_class_object()(),
                                           parameter=parameter, suite_parameter=class_parameter)
            elif test.accepts_suite_parameters():
                test.get_function_object()(suite.get_class_object()(), suite_parameter=class_parameter)
            elif test.accepts_test_parameters():
                test.get_function_object()(suite.get_class_object()(), parameter=parameter)
            else:
                test.get_function_object()(suite.get_class_object()())

    @staticmethod
    def __process_event(event, suite, test=None, param=None, class_param=None, error=None, formatted_traceback=None):

        event_mapping = {Event.ON_SUCCESS: {"custom": suite.get_listener().on_success,
                                            "native": Listener().on_success},
                         Event.ON_FAILURE: {"custom": suite.get_listener().on_failure,
                                            "native": Listener().on_failure},
                         Event.ON_ERROR: {"custom": suite.get_listener().on_error,
                                          "native": Listener().on_error},
                         Event.ON_SKIP: {"custom": suite.get_listener().on_skip,
                                         "native": Listener().on_skip},
                         Event.ON_CANCEL: {"custom": suite.get_listener().on_cancel,
                                           "native": Listener().on_cancel},
                         Event.ON_IGNORE: {"custom": suite.get_listener().on_ignore,
                                           "native": Listener().on_ignore},
                         Event.ON_CLASS_CANCEL: {"custom": suite.get_listener().on_class_cancel,
                                                 "native": Listener().on_class_cancel},
                         Event.ON_CLASS_SKIP: {"custom": suite.get_listener().on_class_skip,
                                               "native": Listener().on_class_skip},
                         Event.ON_BEFORE_CLASS_ERROR: {"custom": suite.get_listener().on_before_class_error,
                                                       "native": Listener().on_before_class_error},
                         Event.ON_BEFORE_CLASS_FAIL: {"custom": suite.get_listener().on_before_class_failure,
                                                      "native": Listener().on_before_class_failure},
                         Event.ON_CLASS_IGNORE: {"custom": suite.get_listener().on_class_ignore,
                                                 "native": Listener().on_class_ignore},
                         Event.ON_AFTER_CLASS_ERROR: {"custom": suite.get_listener().on_after_class_error,
                                                      "native": Listener().on_after_class_error},
                         Event.ON_AFTER_CLASS_FAIL: {"custom": suite.get_listener().on_after_class_failure,
                                                     "native": Listener().on_after_class_failure},
                         Event.ON_CLASS_IN_PROGRESS: {"custom": suite.get_listener().on_class_in_progress,
                                                      "native": Listener().on_class_in_progress},
                         Event.ON_CLASS_COMPLETE: {"custom": suite.get_listener().on_class_complete,
                                                   "native": Listener().on_class_complete}
                         }

        native_function = event_mapping[event]["native"]
        custom_function = event_mapping[event]["custom"]
        custom_function = custom_function \
            if str(custom_function).split(" at ")[0] != str(native_function).split(" at ")[0] else None

        @synchronized(threading.Lock())
        def __create_properties():
            properties = {"suite_meta": suite.get_meta(copy_of_meta=True),
                          "test_meta": test.get_meta(param, class_param, copy_of_meta=True) if test else None,
                          "jm": {"jso": suite}}
            if test:
                properties["test_meta"].update({"parameter": param})
                properties["jm"].update({"jto": test})
            properties["suite_meta"].update({"parameter": class_param})
            return properties
        try:

            if error is not None:
                native_function(custom_function=custom_function,
                                properties=__create_properties(),
                                error=error,
                                trace=formatted_traceback)
            else:
                native_function(custom_function=custom_function,
                                properties=__create_properties())
        except Exception:
            trace = ""
            if sys.version_info[0] < 3:
                trace = "\n\n{}".format(traceback.format_exc())
            raise TestListenerError("Exception occurred while processing custom event listener for function: {}. "
                                    "For help on defining custom event listeners, see documentation: {}{}"
                                    .format(custom_function, DocumentationLinks.LISTENERS, trace))

    @staticmethod
    def __runnable_tags(test, run_config):
        """
        This function evaluates if test should be executed based on its tags and the tag config
        :param test: TestObject
        :param run_config: DICT
        :return: BOOLEAN
        """
        test_tags = test.get_tags()
        tag_config = run_config.get("tag_config", {})

        def __config_set(config):
            return tag_config.get(config, None) is not None

        def __full_match(config):
            for tag in tag_config[config]:
                if tag not in test_tags:
                    return False
            return True

        def __partial_match(config):
            for tag in tag_config[config]:
                if tag in test_tags:
                    return True

        if tag_config is not None:
            try:
                if __config_set("skip_on_match_all") and __full_match("skip_on_match_all"):
                    return False
                elif __config_set("skip_on_match_any") and __partial_match("skip_on_match_any"):
                    return False
                elif __config_set("run_on_match_all") and __full_match("run_on_match_all"):
                    return True
                elif __config_set("run_on_match_any") and __partial_match("run_on_match_any"):
                    return True
            except Exception:
                trace = ""
                if sys.version_info[0] < 3:
                    trace = "\n\n{}".format(traceback.format_exc())
                raise ConfigError("Error occurred while trying to parse `tag_config`. For help on defining the config, "
                                  "see documentation: {}{}".format(DocumentationLinks.TAGS, trace))
        return True

    @staticmethod
    def __positive_skip_condition(test, run_config):
        """
        This function will evaluate inputs for skip parameter whether its a function or a boolean
        :param test: TestObject object
        :param run_config: DICT
        :return: BOOLEAN
        """
        val = test.can_skip()
        components = run_config.get("components", None)
        owners = run_config.get("owners", None)
        tests = run_config.get("tests", None)

        # Run only tests that were requested
        if val is False and tests is not None:
            if sys.version_info[0] < 3:
                """
                While both in Python 2 and 3 function objects look like, tests.junkie_suites.SkipSuites
                in Python 2 the list gets stored as:
                    [<unbound method SkipTests.test_1>, <unbound method SkipTests.test_2>]
                and in Python 3 its stored as:
                    [<function SkipTests.test_1 at 0x0438A780>, <function SkipTests.test_2 at 0x0438A7C8>]
                So oddly enough in Python 2 the regular check did not work, thus this hack
                """
                val = True
                for t in tests:
                    if inspect.getsource(test.get_function_object()) == inspect.getsource(t):
                        val = False
                        break
            else:
                val = not test.get_function_object() in tests

        # Run only components that were requested
        if val is False and components is not None:
            val = not test.get_component() in components

        # Run only tests that belong to the owners requested
        if val is False and owners is not None:
            val = not test.get_owner() in owners
        return val



from test_junkie.decorators import test, afterTest, Suite
from tests.junkie_suites.TestListener import TestListener


@Suite(retry=2, listener=TestListener)
class AfterTestExceptionSuite:

    @afterTest()
    def after_test(self):
        raise Exception("Assertion Error in before class")

    @test(retry=2)
    def ignore_1(self):
        pass

    @test(retry=2)
    def ignore_2(self):
        pass
